import tomli
from .messenger.slack import Slack
from .messenger.discord import Discord
from .messenger.cqhttp import CQHttp
from .hub import Hub
from typing import List


def main():
    with open("bygeon.toml", "rb") as f:
        config = tomli.load(f)

    hub_configs = config["Hubs"]
    hubs: List[Hub] = []

    for (i, hub_config) in enumerate(hub_configs):
        hub_name = hub_config.get("name", f"HUB-{i}")
        keep_data = hub_config["keep_data"]

        hub = Hub(hub_name)
        hubs.append(hub)

        if hub_config.get("Discord") != None:
            discord = Discord(
                bot_token=hub_config["Discord"]["bot_token"],
                channel_id=hub_config["Discord"]["channel_id"],
                hub=hub,
            )
            hub.add_client(discord)
        if hub_config.get("Slack") != None:
            slack = Slack(
                app_token=hub_config["Slack"]["app_token"],
                bot_token=hub_config["Slack"]["bot_token"],
                channel_id=hub_config["Slack"]["channel_id"],
                hub=hub,
            )
            hub.add_client(slack)
        if hub_config.get("CQHttp") != None:

            ws_url = hub_config["CQHttp"].get("ws_url", "ws://localhost:8080/")
            http_url = hub_config["CQHttp"].get("ws_url", "http://localhost:5700/")

            cqhttp = CQHttp(
                group_id=hub_config["CQHttp"]["group_id"],
                hub=hub,
                ws_url=ws_url,
                http_url=http_url,
            )

            hub.add_client(cqhttp)

        hub.init_database(keep_data=keep_data)
        hub.start()

    for hub in hubs:
        hub.join()



from typing import List
from pypika import Query, Column, Table
import sqlite3

from .messenger.messenger import Messenger
from .message import Message

import bygeon.util as util


class Hub:
    def __init__(self, name: str) -> None:
        self.clients: List[Messenger] = []
        self.con = sqlite3.connect(f"{name}.db", check_same_thread=False)
        self.cur = self.con.cursor()
        self.name = name

    @property
    def client_names(self) -> List[str]:
        return [client.name for client in self.clients]

    def start(self):
        for client in self.clients:
            client.start()

    def join(self):
        for client in self.clients:
            client.join()

    def new_message(self, message: Message):
        self.new_entry(message)
        for client in self.clients:
            if client.name != message.origin:
                util.run_in_thread(client.send_message, (message,))

    def new_entry(self, message: Message) -> None:
        origin_id = message.origin_id
        origin = message.origin
        entry = (origin_id if origin == s else None for s in self.client_names)
        q = Query.into(self.table).insert(*tuple(entry))
        self.execute_sql(str(q))

    # FIXME
    def update_entry(
        self, m: Message, sent_messenger: str, sent_id: str
    ) -> None:  # noqa
        sql = f'UPDATE "messages" SET "{sent_messenger}" = \'{sent_id}\' WHERE "{m.origin}" = \'{m.origin_id}\''  # noqa
        # q = Query.update(self.table).set(sent_messenger, sent_id).where(m.origin == m.origin_id) # noqa
        self.execute_sql(sql)

    def add_client(self, client):
        self.clients.append(client)

    def reply_message(self, m: Message, reply_to: str) -> None:
        self.new_entry(m)
        orig = m.origin
        sql = f'SELECT * FROM "messages" WHERE "{orig}" = \'{reply_to}\''
        cur = self.cur.execute(sql)

        for row in cur:
            for i, client in enumerate(self.clients):
                if client.name != orig:
                    util.run_in_thread(client.send_message, (m, row[i]))
        print(self.cur.execute(sql))

    def modify_message(self, m: Message) -> None:
        orig = m.origin
        sent_id = m.origin_id
        sql = f'SELECT * FROM "messages" WHERE "{orig}" = \'{sent_id}\''
        cur = self.cur.execute(sql)
        for row in cur:
            for i, client in enumerate(self.clients):
                if client.name != orig:
                    util.run_in_thread(client.modify_message, (m, row[i]))
        ...

    def recall_message(self, orig: str, recalled_id: str) -> None:
        sql = f'SELECT * FROM "messages" WHERE "{orig}" = \'{recalled_id}\''
        cur = self.cur.execute(sql)
        for row in cur:
            for i, client in enumerate(self.clients):
                if client.name != orig:
                    util.run_in_thread(client.recall_message, (row[i],))

    def init_database(self, keep_data=True):
        columns = tuple(
            Column(s, "VARCHAR(255)", nullable=True) for s in self.client_names
        )
        if not keep_data:
            self.execute_sql('DROP TABLE IF EXISTS "messages"')
            create_table = Query.create_table("messages").columns(*columns)
            self.execute_sql(str(create_table))
        self.table = Table("messages")

    def execute_sql(self, query: str) -> None:
        print(query)
        self.cur.execute(query)
        self.con.commit()



import threading
from typing import Union, cast
from urllib.parse import urljoin

from websocket import WebSocketApp as WSApp

import orjson
import requests

import bygeon.util as util
from bygeon.hub import Hub
from bygeon.message import Message, Attachment
from .definition.cqhttp import WSMessage, PostType, Endpoints
from .messenger import Messenger


class CQHttp(Messenger):
    @property
    def send_url(self) -> str:
        return urljoin(self.http_url, Endpoints.SEND_GROUP_MESSAGE)

    @property
    def recall_url(self) -> str:
        return urljoin(self.http_url, Endpoints.DELETE_MESSAGE)

    def __init__(self, group_id: str, hub: Hub, ws_url: str, http_url: str) -> None:
        self.group_id = int(group_id)
        self.hub = hub
        self.logger = self.get_logger()
        self.ws_url = ws_url
        self.http_url = http_url

    def on_open(self, ws) -> None:
        self._on_open(ws)

    def on_error(self, ws, e) -> None:
        self._on_error(ws, e)

    def on_close(self, ws, close_status_code, close_msg) -> None:
        self._on_close(ws, close_status_code, close_msg)

    def on_message(self, ws: WSApp, message: str) -> None:
        ws_message: WSMessage = orjson.loads(message)
        post_type = ws_message["post_type"]
        is_reply = False
        message_group_id = ws_message.get("group_id")
        match post_type:
            case PostType.MESSAGE:
                if message_group_id != self.group_id:
                    return None
                if ws_message["self_id"] == ws_message["user_id"]:
                    return None
                message_id = ws_message["message_id"]

                self.logger.info("Received message: " + str(message_id))

                author = ws_message["sender"]["nickname"]

                data = ws_message["message"]
                text = ""
                attachments = []
                for d in data:
                    if d["type"] == "reply":
                        is_reply = True
                        ref_id = d["data"]["id"]

                        self.logger.info("Reply to: " + str(ref_id))
                    elif d["type"] == "text":
                        text += d["data"]["text"]
                    elif d["type"] == "image":
                        url = d["data"].get("url", "")
                        url = cast(str, url)
                        fn = d["data"]["file"]
                        filename = f"{self.name}_{fn}"
                        path = self.generate_cache_path(self.hub.name)
                        file_path = util.download_to_cache(url, path, filename)
                        attachments.append(Attachment(fn, "image", file_path))
                m = Message(self.name, message_id, author, text, attachments)
                if is_reply:
                    self.hub.reply_message(m, ref_id)
                else:
                    self.hub.new_message(m)
            case PostType.NOTICE:
                if ws_message["self_id"] == ws_message["user_id"]:
                    return None
                recalled_id = ws_message["message_id"]
                self.hub.recall_message(self.name, recalled_id)

    def recall_message(self, message_id: str) -> None:
        payload = {
            "message_id": message_id,
        }
        r = requests.post(self.recall_url, json=payload)
        self.logger.info("Trying to recall: " + message_id)
        self.logger.info(r.json())

    def modify_message(self, m: Message, m_id: str) -> None:
        self.recall_message(m_id)

        self.send_message(m)
        ...

    def reconnect(self) -> None:
        ...

    def send_message(self, m: Message, ref_id=None) -> None:
        payload: dict[str, Union[str, int]] = {
            "group_id": self.group_id,
            "message": "",
        }
        message_string = ""
        for attachment in m.attachments:
            main_type = attachment.type.split("/")[0]
            message_string += f"[CQ:{main_type},file=file:{attachment.file_path}]"

        self.logger.info(ref_id)
        if ref_id is not None:
            message_string += f"[CQ:reply,id={ref_id}]"
        message_string += f"[{m.author_username}]: {m.text}"
        self.logger.info(f"Sending message with CQCode: {message_string}")
        payload["message"] = message_string
        self.logger.info(payload)

        r = requests.post(self.send_url, json=payload)
        self.logger.info(r.text)

        response = r.json()
        message_id = response.get("data").get("message_id")
        self.hub.update_entry(m, self.name, message_id)

    def start(self) -> None:
        self.ws = WSApp(
            self.ws_url,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.thread = threading.Thread(target=self.ws.run_forever)
        self.thread.daemon = True
        self.thread.start()

    def join(self) -> None:
        self.thread.join()



import threading
from os.path import basename
from typing import cast, List

from websocket import WebSocketApp as WSApp

import requests
import orjson

import bygeon.util as util
from bygeon.hub import Hub
from bygeon.message import Message, Attachment
from .messenger import Messenger
from .definition.slack import WSMessageType, EventType, MessageEventSubtype
from .definition.slack import Endpoints, WSMessage, Event, MessageEvent, File


class Slack(Messenger):
    def __init__(
        self, app_token: str, bot_token: str, channel_id: str, hub: Hub
    ) -> None:

        self.app_token = app_token
        self.bot_token = bot_token
        self.channel_id = channel_id
        self.hub = hub
        self.logger = self.get_logger()

        self.bot_user_id = self.get_bot_user_id()

    def on_open(self, ws) -> None:
        self._on_open(ws)

    def on_error(self, ws, e) -> None:
        self._on_error(ws, e)

    def on_close(self, ws, close_status_code, close_msg) -> None:
        self._on_close(ws, close_status_code, close_msg)

    def on_message(self, ws: WSApp, message: str) -> None:
        self.logger.debug(message)
        ws_message: WSMessage = orjson.loads(message)
        ws_type = ws_message["type"]

        match ws_type:
            case WSMessageType.HELLO:
                return None
            # FIXME
            case WSMessageType.DISCONNECT:
                self.logger.error("Disconnected")
                self.logger.error("Trying to reconnect")
                self.reconnect()
            case WSMessageType.EVENTS_API:
                event = ws_message["payload"]["event"]
                self.send_ack(ws, ws_message)
                self.handle_event(event)

    def handle_event(self, event: Event) -> None:
        event_type = event["type"]
        match event_type:
            case EventType.MESSAGE:
                event = cast(MessageEvent, event)
                self.handle_message(event)
            case _:
                return None

    def handle_message(self, event: MessageEvent) -> None:
        # XXX
        subtype = event.get("subtype", "no_subtype")
        message_id = event["ts"]
        text = event["text"]
        user_id = event.get("user")

        if user_id is None:
            username = event.get("username", "")
        else:
            username = self.get_username(user_id)

        if user_id == self.bot_user_id:
            return None

        # XXX
        username = cast(str, username)

        if event["channel"] != self.channel_id:
            return None

        match subtype:
            case MessageEventSubtype.MESSAGE_DELETED:
                deleted_ts = event["deleted_ts"]
                self.logger.info("Deleted message: {}".format(deleted_ts))
                self.hub.recall_message(self.name, deleted_ts)

            case MessageEventSubtype.BOT_MESSAGE:
                ...

            case MessageEventSubtype.NO_SUBTYPE:

                m = Message(self.name, message_id, username, text, [])
                if (ref_id := event.get("thread_ts")) is not None:
                    self.hub.reply_message(m, ref_id)
                else:
                    self.hub.new_message(m)
            case MessageEventSubtype.FILE_SHARE:
                attachments = self.get_attachments(event)
                m = Message(self.name, message_id, username, text, attachments)
                self.hub.new_message(m)
            case MessageEventSubtype.MESSAGE_CHANGED:
                self.logger.info(event)
                m = Message(self.name, message_id, username, text, [])
                self.hub.modify_message(m)

    def get_attachments(self, event) -> list:
        files: List[File] = event.get("files", [])
        attachment = []
        for file in files:
            fn = self.cache_prefix(file["id"]) + file["name"]
            self.logger.info("Downloading file: {}".format(fn))
            url = file["url_private_download"]
            t = file["mimetype"]
            path = self.generate_cache_path(self.hub.name)
            file_path = util.download_to_cache(
                url, path, fn, headers=self.get_headers(self.bot_token)
            )
            a = Attachment(fn, t, file_path)
            attachment.append(a)
        return attachment

    def send_ack(self, ws: WSApp, message: WSMessage) -> None:
        envelope_id = message["envelope_id"]
        ws.send(orjson.dumps({"envelope_id": envelope_id}))

    def get_username(self, id: str) -> str:
        headers = self.get_headers(self.bot_token)
        r = requests.get(Endpoints.USERS_INFO + "?user=" + id, headers=headers)
        response = orjson.loads(r.text)
        self.logger.debug(r.text)
        username = response["user"]["name"]
        return username

    def reconnect(self) -> None:
        self.ws.close()
        self.start()
        self.join()

    def get_websocket_url(self) -> str:
        header = self.get_headers(self.app_token)
        r = requests.post(Endpoints.CONNECTIONS_OPEN, headers=header)
        response = orjson.loads(r.text)
        self.logger.debug(response)

        try:
            websocket_url = response["url"]
        except KeyError:
            self.logger.error("Could not get websocket url")
            raise Exception("Could not get websocket url")
        else:
            self.logger.info("Successfully got websocket url")

        return websocket_url

    def send_message(self, m: Message, ref_id=None) -> None:
        payload = {
            "type": "message",
            "username": m.author_username,
            "channel": self.channel_id,
            "text": m.text,
        }
        if ref_id is not None:
            payload["thread_ts"] = ref_id
        if len(m.attachments) != 0:
            payload["initial_comment"] = m.text
            self.upload_files(m)
        self.logger.info("Sending message: {}".format(m.text))
        r = requests.post(
            Endpoints.POST_MESSAGE,
            json=payload,
            headers=self.get_headers(self.bot_token),
        )
        response = orjson.loads(r.text)
        if not response["ok"]:
            self.logger.error(response)
        else:
            self.hub.update_entry(m, self.name, response.get("ts"))

    # return last id as message id
    def upload_files(self, m: Message) -> None:
        payload = {"channels": self.channel_id}

        attachments = m.attachments

        headers = self.get_headers(self.bot_token)
        headers.pop("Content-Type")
        for attachment in attachments:
            fn = basename(attachment.file_path)
            a_type = attachment.type
            file = {"file": (fn, open(attachment.file_path, "rb"), a_type)}
            self.logger.info(attachment.file_path)
            r = requests.post(
                Endpoints.FILE_UPLOAD,
                headers=headers,
                data=payload,
                files=file,
            )
            response = orjson.loads(r.text)
            if not response["ok"]:
                self.logger.error(response)

    def recall_message(self, message_id: str) -> None:
        payload = {
            "token": self.bot_token,
            "channel": self.channel_id,
            "ts": message_id,
        }
        r = requests.post(
            Endpoints.CHAT_DELETE,
            json=payload,
            headers=self.get_headers(self.bot_token),
        )
        self.logger.info("Trying to recall: " + message_id)

    def modify_message(self, m: Message, m_id: str) -> None:
        payload = {
            "token": self.bot_token,
            "channel": self.channel_id,
            "ts": m_id,
            "text": m.text,
        }
        requests.post(
            Endpoints.CHAT_UPDATE,
            json=payload,
            headers=self.get_headers(self.bot_token),
        )

    def start(self) -> None:
        self.ws = WSApp(
            self.get_websocket_url(),
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.thread = threading.Thread(target=self.ws.run_forever)
        self.thread.daemon = True
        self.thread.start()

    def get_bot_user_id(self) -> str:
        headers = self.get_headers(self.bot_token)

        r = requests.get(Endpoints.AUTH_TEST, headers=headers)
        bot_info = orjson.loads(r.text)
        return bot_info["user_id"]

    def get_headers(self, token) -> dict:
        return {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + token,
        }

    def join(self) -> None:
        self.thread.join()



from typing import Union, List, Optional, TypedDict
from typing_extensions import NotRequired


class Endpoints:
    GATEWAY = "wss://gateway.discord.gg/?v=10&encoding=json"
    SEND_MESSAGE = "https://discordapp.com/api/channels/{}/messages"
    DELETE_MESSAGE = "https://discordapp.com/api/channels/{}/messages/{}"
    EDIT_MESSAGE = "https://discordapp.com/api/channels/{}/messages/{}"
    GET_EMOJI = "https://cdn.discordapp.com/emojis/{}"


class ReferencedMessage(TypedDict):
    type: int
    id: str


class Author(TypedDict):
    id: str
    username: str
    avatar: str
    bot: bool


class EmbedAuthor(TypedDict):
    name: str
    url: str
    icon_url: str
    proxy_icon_url: str


class GatewayEvent(TypedDict):
    type: int
    referenced_message: ReferencedMessage
    channel_id: str
    content: str
    author: Author
    heartbeat_interval: int


class UnavailableGuild(TypedDict):
    id: str
    unavailable: bool


class User(TypedDict):
    id: str
    username: str
    discriminator: str
    avatar: Optional[str]
    bot: NotRequired[bool]
    system: NotRequired[bool]
    mfa_enabled: NotRequired[bool]
    banner: NotRequired[Optional[str]]
    accent_color: NotRequired[Optional[int]]
    locale: NotRequired[str]
    verified: NotRequired[bool]
    email: NotRequired[Optional[str]]
    flags: NotRequired[int]
    premium_type: NotRequired[int]
    public_flags: NotRequired[int]


class Role(TypedDict):
    pass


class Attachment(TypedDict):
    id: str
    filename: str
    description: NotRequired[str]
    content_type: NotRequired[str]
    # size	integer	size of file in bytes
    size: int
    url: str
    proxy_url: str
    height: NotRequired[Optional[int]]
    width: NotRequired[Optional[int]]
    ephemeral: NotRequired[bool]


class Embed(TypedDict):
    pass


class Reaction(TypedDict):
    pass


class MessageReference(TypedDict):
    pass


class Channel(TypedDict):
    pass


class StickerItem(TypedDict):
    id: str
    name: str
    # 1: png, 2: apng, 3: lottie
    format_type: int


class DiscordMessage(TypedDict):
    id: str
    channel_id: str
    author: User
    content: str
    timestamp: str
    edited_timestamp: Optional[str]
    tts: bool
    mention_everyone: bool
    mentions: List[User]
    mention_roles: List[Role]
    # mention_channels?**	array of channel mention objects	channels specifically mentioned in this message
    attachments: List[Attachment]
    embeds: List[Embed]
    reactions: Optional[List[Reaction]]
    nonce: NotRequired[Union[int, str]]
    pinned: bool
    webhook_id: NotRequired[str]
    type: int
    # activity?	message activity object	sent with Rich Presence-related chat embeds
    # application?	partial application object	sent with Rich Presence-related chat embeds
    # application_id?	snowflake	if the message is an Interaction or application-owned webhook, this is the id of the application
    message_reference: NotRequired[MessageReference]
    flags: NotRequired[int]

    # recursive TypedDict not supported yet
    referenced_message: NotRequired[Optional[dict]]
    thread: NotRequired[Channel]

    # components?	Array of message components	sent if the message contains components like buttons, action rows, or other interactive components
    sticker_items: Optional[List[StickerItem]]


class MessageCreateEvent(DiscordMessage):
    guild_id: NotRequired[str]
    # mentions: List[User]


class MessageDeleteEvent(TypedDict):
    id: str
    channel_id: str
    guild_id: NotRequired[str]


class MessageUpdateEvent(MessageCreateEvent):
    ...


class Hello(TypedDict):
    heartbeat_interval: int


class Application(TypedDict):
    id: str
    flags: NotRequired[int]


class ReadyEvent(TypedDict):
    v: int
    user: User
    guilds: List[UnavailableGuild]
    session_id: str
    shard: Optional[List[int]]
    application: Application


class WebsocketMessage(TypedDict):
    op: int
    t: str
    s: int
    d: Union[MessageCreateEvent, ReadyEvent, Hello, MessageDeleteEvent]


class Opcode:
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE_UPDATE = 3
    RESUME = 6
    RECONNECT = 7
    HELLO = 10
    HEARTBEAT_ACK = 11


class EventName:
    MESSAGE_CREATE = "MESSAGE_CREATE"
    MESSAGE_UPDATE = "MESSAGE_UPDATE"
    MESSAGE_DELETE = "MESSAGE_DELETE"
    READY = "READY"



from typing import List, TypedDict
from typing_extensions import NotRequired


class Endpoints:
    POST_MESSAGE = "https://slack.com/api/chat.postMessage"
    USERS_INFO = "https://slack.com/api/users.info"
    CONNECTIONS_OPEN = "https://slack.com/api/apps.connections.open"
    CHAT_DELETE = "https://slack.com/api/chat.delete"
    BOTS_INFO = "https://slack.com/api/bots.info"
    AUTH_TEST = "https://slack.com/api/auth.test"
    FILE_UPLOAD = "https://slack.com/api/files.upload"
    CHAT_UPDATE = "https://slack.com/api/chat.update"


class WSMessageType:
    HELLO = "hello"
    DISCONNECT = "disconnect"
    EVENTS_API = "events_api"


class EventType:
    MESSAGE = "message"


class MessageEventSubtype:
    BOT_MESSAGE = "bot_message"
    MESSAGE_CHANGED = "message_changed"
    MESSAGE_DELETED = "message_deleted"

    # Non-existent in WebSocket event
    MESSAGE_REPLIED = "message_replied"
    NO_SUBTYPE = "no_subtype"
    FILE_SHARE = "file_share"


class File(TypedDict):
    id: str
    mimetype: str
    name: str
    title: str
    url_private_download: str


class Event(TypedDict):
    type: str
    subtype: NotRequired[str]


class MessageEvent(Event):
    channel: str
    user: str
    text: NotRequired[str]
    deleted_ts: NotRequired[str]
    thread_ts: NotRequired[str]

    # Use this to create a reply thread
    ts: str
    files: NotRequired[List[File]]


class Element(TypedDict):
    type: str
    text: str


class Block(TypedDict):
    type: str
    block_id: str
    elements: List[Element]


class TeamJoinEvent:
    pass


class PinAddedEvent(TypedDict):
    pass


class UserProfileChangedEvent(TypedDict):
    pass


class Payload(TypedDict):
    token: str
    team_id: str
    event: Event
    client_msg_id: str
    type: str
    text: str
    user: str
    # use ts to create new reply thread
    ts: str
    team: str
    blocks: Block


class WSMessage(TypedDict):
    envelope_id: str
    payload: Payload
    type: str



from threading import Thread
from typing import Callable
import requests
import os
from pathlib import Path


def run_in_thread(func: Callable, args: tuple):
    thread = Thread(target=func, args=args)
    thread.start()


def download_to_cache(url: str, directory: str, filename: str, headers=None):
    Path(directory).mkdir(parents=True, exist_ok=True)

    with requests.get(url, stream=True, headers=headers) as r:
        r.raise_for_status()
        content_type = r.headers["content-type"]

        filename = rename_with_proper_suffix(filename, content_type)

        file_path = os.path.join(directory, filename)
        with open(file_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
        return file_path


def rename_with_proper_suffix(filename: str, content_type: str) -> str:
    suffix = "." + content_type.split("/")[1]
    if not filename.endswith(suffix):
        filename += suffix
    return filename



import threading
import time
import re
from io import BytesIO
from os.path import basename
from typing import cast, List, Dict, Any, Union, Optional

from websocket import WebSocketApp as WSApp

import requests
import orjson

import bygeon.util as util
from bygeon.hub import Hub
from bygeon.message import Message, Attachment
from .messenger import Messenger
from .definition.discord import (
    MessageUpdateEvent,
    Opcode,
    EventName,
    WebsocketMessage,
    Endpoints,
)
from .definition.discord import (
    MessageCreateEvent,
    ReadyEvent,
    Hello,
    MessageDeleteEvent,
)


class Discord(Messenger):
    session_id: Optional[str]
    sequence: Optional[int]

    def __init__(self, bot_token: str, channel_id: str, hub: Hub) -> None:
        self.token = bot_token
        self.channel_id = channel_id
        self.hub = hub
        self.sequence = None
        self.session_id = None

        self.logger = self.get_logger()

    @property
    def headers(self):
        return {"Authorization": f"Bot {self.token}"}

    def on_open(self, ws) -> None:
        self._on_open(ws)

    def on_error(self, ws, e) -> None:
        self._on_error(ws, e)

    def on_close(self, ws, close_status_code, close_msg) -> None:
        self._on_close(ws, close_status_code, close_msg)
        self.reconnect()

    def heartbeat(self, ws: WSApp, interval: int) -> None:
        payload = {
            "op": 1,
            "d": None,
        }
        while True:
            time.sleep(interval / 1000)
            if ws.sock is not None:
                ws.send(orjson.dumps(payload))
            else:
                break

    def on_message(self, ws: WSApp, message: str) -> None:

        ws_message: WebsocketMessage = orjson.loads(message)
        opcode = ws_message["op"]

        match opcode:
            case Opcode.HELLO:
                hello = cast(Hello, ws_message["d"])
                heartbeat_interval = hello["heartbeat_interval"]
                self.send_identity(ws)
                threading.Thread(
                    target=self.heartbeat, args=(ws, heartbeat_interval), daemon=True
                ).start()
            case Opcode.HEARTBEAT:
                # TODO
                pass
            case Opcode.DISPATCH:
                self.logger.debug(ws_message)
                self.handle_dispatch(ws_message)
            case _:
                return None

    def handle_dispatch(self, ws_message: WebsocketMessage) -> None:
        t = ws_message["t"]
        self.sequence = ws_message["s"]

        match t:
            case EventName.MESSAGE_CREATE:
                create_event = cast(MessageCreateEvent, ws_message["d"])
                self.handle_message_create(create_event)
            case EventName.MESSAGE_DELETE:
                delete_event = cast(MessageDeleteEvent, ws_message["d"])
                message_id = delete_event["id"]
                self.hub.recall_message(self.name, message_id)
            case EventName.READY:
                ready_event = cast(ReadyEvent, ws_message["d"])
                self.handle_ready(ready_event)
            case EventName.MESSAGE_UPDATE:
                update_event = cast(MessageUpdateEvent, ws_message["d"])
                text = update_event["content"]
                username = update_event["author"]["username"]
                message_id = update_event["id"]
                m = Message(self.name, message_id, username, text, [])
                self.hub.modify_message(m)
            case _:
                return None

    def handle_ready(self, data: ReadyEvent) -> None:
        self.bot_id = data["user"]["id"]
        self.session_id = data["session_id"]

    def handle_reply(self, m: Message, ref_id: str) -> None:
        self.hub.reply_message(m, ref_id)

    def handle_modify(self, data: MessageUpdateEvent) -> None:
        message_id = data["id"]
        author = data["author"]
        username = author["username"]
        text = data["content"]
        m = Message(self.name, message_id, username, text, [])
        self.hub.modify_message(m)

    def modify_message(self, m: Message, m_id: str) -> None:
        url = Endpoints.EDIT_MESSAGE.format(self.channel_id, m_id)

        payload = {
            "content": f"[{m.author_username}]: {m.text}",
        }

        requests.patch(url, headers=self.headers, json=payload)

    def handle_message_create(self, data: MessageCreateEvent) -> None:
        if data.get("channel_id") != self.channel_id:
            return None
        elif data["author"].get("id") == self.bot_id:
            return None

        origin_id = data["id"]

        text = data["content"]
        self.logger.info("Received message: %s", text)

        author = data["author"]
        username = author["username"]
        attachments: List[Attachment] = []
        for attachment in data["attachments"]:
            url = attachment["url"]

            fn = attachment["filename"]
            filename = f"{self.name}_{fn}"

            full_type = attachment["content_type"]

            path = self.generate_cache_path(self.hub.name)
            file_path = util.download_to_cache(url, path, filename)
            attachments.append(Attachment(fn, full_type, file_path))

        emoji_regex = r"<:(.+):(\d+)>"
        emoji_re = re.compile(emoji_regex)
        emoji_list = emoji_re.findall(text)

        a_emoji_regex = r"<a:(.+):(\d+)>"
        a_emoji_re = re.compile(a_emoji_regex)
        a_emoji_list = a_emoji_re.findall(text)
        for a_emoji_name, a_emoji_id in a_emoji_list:
            fn = f"{a_emoji_name}_{a_emoji_id}.gif"
            url = Endpoints.GET_EMOJI.format(a_emoji_id) + ".gif"
            path = self.generate_cache_path(self.hub.name)
            file_path = util.download_to_cache(url, path, fn)
            full_type = "image/gif"
            attachments.append(Attachment(fn, full_type, file_path))
            text = text.replace(f"<a:{a_emoji_name}:{a_emoji_id}>", "")

        for emoji_name, emoji_id in emoji_list:
            fn = f"{emoji_name}_{emoji_id}.png"
            url = Endpoints.GET_EMOJI.format(emoji_id) + ".png"
            path = self.generate_cache_path(self.hub.name)
            file_path = util.download_to_cache(url, path, fn)
            full_type = "image/png"
            attachments.append(Attachment(fn, full_type, file_path))
            text = text.replace(f"<:{emoji_name}:{emoji_id}>", "")

        if (sticker_items := data.get("sticker_items")) is not None:
            for sticker in sticker_items:
                fn = sticker["id"]
                match sticker["format_type"]:
                    case 1:
                        fn += ".png"
                        full_type = "image/png"
                    case 2:
                        fn += ".apng"
                        full_type = "image/apng"
                    case 3:
                        continue
                        # fn += ".lottie"
                        # full_type = "application/json"
                path = self.generate_cache_path(self.hub.name)
                file_path = util.download_to_cache(url, path, filename)
                attachments.append(Attachment(fn, full_type, file_path))

        m = Message(self.name, origin_id, username, text, attachments)
        if (ref_message := data["referenced_message"]) is not None:
            ref_id = ref_message["id"]
            self.hub.reply_message(m, ref_id)
        else:
            self.hub.new_message(m)

    def recall_message(self, message_id: str) -> None:
        r = requests.delete(
            Endpoints.DELETE_MESSAGE.format(self.channel_id, message_id),
            headers=self.headers,
        )
        self.log_response(r)

    def send_message(self, m: Message, ref_id=None) -> None:
        payload: dict[str, Union[str, dict]] = {
            "content": f"[{m.author_username}]: {m.text}"
        }
        if ref_id is not None:
            payload["message_reference"] = {
                "channel_id": self.channel_id,
                "message_id": ref_id,
            }

        files = []
        for (i, attachment) in enumerate(m.attachments):
            fn = basename(attachment.file_path)
            a_type = attachment.type
            files.append(
                (f"files[{i}]", (fn, open(attachment.file_path, "rb"), a_type))
            )

        if len(files) > 0:
            # XXX
            payload_io = BytesIO(orjson.dumps(payload))
            files.append(
                (
                    "payload_json",
                    (None, payload_io, "application/json"),
                )  # type: ignore[arg-type]
            )
            r = requests.post(
                Endpoints.SEND_MESSAGE.format(self.channel_id),
                headers=self.headers,
                files=files,
            )
        else:
            r = requests.post(
                Endpoints.SEND_MESSAGE.format(self.channel_id),
                json=payload,
                headers=self.headers,
            )

        self.log_response(r)

        message_id: str = r.json().get("id")
        self.hub.update_entry(m, self.name, message_id)

    def send_identity(self, ws: WSApp) -> None:
        payload = self.identity_payload
        print(payload)
        ws.send(payload)

    @property
    def identity_payload(self) -> bytes:
        # XXX
        payload: Dict[str, Union[dict, Any]] = {
            "op": Opcode.IDENTIFY,
            "d": {
                "token": self.token,
                "properties": {
                    "os": "linux",
                    "browser": "bygeon",
                    "device": "bygeon",
                },
                "large_threshold": 250,
                "compress": False,
                "intents": (1 << 15) + (1 << 9),
            },
        }

        if self.sequence is not None:
            payload["sequence"] = self.sequence
        if self.session_id is not None:
            payload["d"]["session_id"] = self.session_id

        return orjson.dumps(payload)

    def log_response(self, r: requests.Response) -> None:
        if r.status_code != 200:
            self.logger.error(r.text)
        else:
            self.logger.debug(r.text)

    def reconnect(self) -> None:
        # XXX
        self.ws.close()
        self.start()
        self.join()

    def start(self) -> None:
        self.ws = WSApp(
            Endpoints.GATEWAY,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.thread = threading.Thread(target=self.ws.run_forever)
        self.thread.daemon = True
        self.thread.start()

    def join(self) -> None:
        self.thread.join()



import os
import colorlog as cl
import logging
from typing import Protocol

from websocket import WebSocketApp as WSApp

from bygeon.message import Message

logger_format = "%(log_color)s%(levelname)s: %(name)s: %(message)s"


class Messenger(Protocol):
    logger: logging.Logger

    def get_logger(self):
        handler = cl.StreamHandler()
        handler.setFormatter(cl.ColoredFormatter(logger_format))

        logger = cl.getLogger(self.name)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        return logger

    @property
    def file_cache_path(self) -> str:
        return os.path.join(os.getcwd(), "cache")

    def generate_cache_path(self, hub_name: str) -> str:
        return os.path.join(self.file_cache_path, hub_name)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def _on_open(self, ws) -> None:
        self.logger.info("Opened WebSocket connection")

    def _on_error(self, ws, e) -> None:
        self.logger.error(f"WebSocket encountered error: {e}")

    def _on_close(self, ws, close_status_code, close_msg) -> None:
        self.logger.error(f"WebSocket closed: {close_msg}")

    def on_message(self, ws: WSApp, message: str) -> None:
        ...

    def send_message(self, m: Message, ref_id=None) -> None:
        ...

    def modify_message(self, m: Message, m_id: str) -> None:
        ...

    def recall_message(self, message_id: str) -> None:
        ...

    def start(self) -> None:
        ...

    def join(self) -> None:
        ...

    def cache_prefix(self, id="") -> str:
        return f"{self.name}_{id}."



import os
import sys

if not __package__:
    source_path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, source_path)

from bygeon.main import main

if __name__ == "__main__":
    main()



from typing import TypedDict, List
from typing_extensions import NotRequired


class Endpoints:
    SEND_GROUP_MESSAGE = "send_group_msg"
    DELETE_MESSAGE = "delete_msg"


class PostType:
    META_EVENT = "meta_event"
    MESSAGE = "message"
    NOTICE = "notice"


class CQType:
    TEXT = "text"
    IMAGE = "image"
    RECORD = "record"
    VIDEO = "video"


ATTACHMENT_TYPES = [CQType.IMAGE, CQType.VIDEO, CQType.RECORD]


class MetaEventType:
    HEARTBEAT = "heartbeat"


class MessageType:
    GROUP = "group"


class Sender(TypedDict):
    nickname: str
    user_id: int


class CQData(TypedDict):
    id: str
    text: NotRequired[str]
    file: NotRequired[str]


class CQMessage(TypedDict):
    type: str
    data: CQData


class WSMessage(TypedDict):
    post_type: str
    meta_event_type: str
    message_type: str
    sender: Sender
    message_id: str
    message: List[CQMessage]
    self_id: NotRequired[int]
    user_id: NotRequired[int]



from typing import NamedTuple, List
from enum import Enum

class AttachmentType(Enum):
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"

class Attachment(NamedTuple):
    name: str
    type: str
    file_path: str

class Message(NamedTuple):
    origin: str
    origin_id: str
    author_username: str
    text: str
    attachments: List[Attachment]
